import logging
from optparse import OptionParser

from twoline.manager import Manager


def run_from_cmdline():
    parser = OptionParser()
    parser.add_option(
        '--port', '-p', dest='port', default='6224'
    )
    parser.add_option(
        '--ip', '-i', dest='ip', default='0.0.0.0'
    )
    parser.add_option(
        '--loglevel', '-l', dest='loglevel', default='INFO'
    )
    parser.add_option(
        '--logpath', dest='logpath', default=None
    )
    parser.add_option(
        '--size-x', '-x', dest='size_x', default='16',
    )
    parser.add_option(
        '--size-y', '-y', dest='size_y', default='2',
    )
    parser.add_option(
        '--text-cycle-interval', dest='text_cycle_interval', default='2'
    )
    parser.add_option(
        '--blink-interval', dest='blink_interval', default='0.25'
    )

    options, args = parser.parse_args()

    if len(args) != 1:
        parser.error('Device required (usually a path in /dev/)')

    if options.logpath:
        lp_logger = logging.getLogger(options.logpath)
        lp_logger.setLevel(options.loglevel)
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(options.loglevel)
        lp_logger.addHandler(stream_handler)
    else:
        logging.basicConfig(
            level=logging.getLevelName(options.loglevel)
        )

    manager = Manager(*args, **vars(options))
    manager.run()



from functools import wraps
import logging
import time


logger = logging.getLogger(__name__)


COMMANDS = {}


def command(fn):
    @wraps(fn)
    def wrapped(*args):
        self = args[0]
        logger.debug(
            'Executing %s%s',
            fn.func_name,
            args
        )
        response = fn(*args)
        logger.debug(
            'Response %s',
            response
        )
        if response:
            self.send_manager_data('response', response)
    COMMANDS[fn.func_name] = wrapped
    return fn


class LcdManager(object):
    def __init__(self, device_path, pipe=None, size=None,
            blink_interval=0.25, text_cycle_interval=2, size_x=16, size_y=2):
        self.device_path = device_path
        self.pipe = pipe
        if not size:
            size = [16, 2]
        self.size = size

        self.message = ''
        self.message_lines = []
        self.color = 0, 0, 0

        self.sleep = 0.1

        self.blink = []
        self.blink_idx = 0
        self.blink_counter = 0
        self.blink_interval = int(
            (1.0 / self.sleep) * blink_interval
        )

        self.text_idx = 0
        self.text_cycle_counter = 0
        self.text_cycle_interval = int(
            (1.0 / self.sleep) * text_cycle_interval
        )

    def initialize(self):
        self.send('\xfe\x52')
        self.clear()

    def run(self):
        while True:
            if self.pipe.poll():
                cmd, args = self.pipe.recv()
                args.insert(0, self)
                if cmd in COMMANDS:
                    COMMANDS[cmd](*args)
                else:
                    logger.error(
                        'Received unknown command \'%s\' from manager.',
                        cmd
                    )
                    self.send_manager_data(
                        'error', 'Command %s does not exist' % cmd
                    )
            if self.blink_counter >= self.blink_interval:
                self.blink_counter = 0
                self.handle_blink()
            else:
                self.blink_counter += 1
            if self.text_cycle_counter >= self.text_cycle_interval:
                self.text_cycle_counter = 0
                self.handle_text_cycle()
            else:
                self.text_cycle_counter += 1
            time.sleep(self.sleep)

    def handle_text_cycle(self):
        if len(self.message_lines) <= self.text_idx:
            self.text_idx = 0
        self.send('\xfe\x58')
        self.send('\xfe\x4700')
        cleaned_lines = [
            line.ljust(self.size[0])
            for line in self.message_lines[
                self.text_idx:self.text_idx+self.size[1]
            ]
        ]
        display_text = ''.join(cleaned_lines)[0:self.size[0]*self.size[1]-1]
        self.send(display_text.encode('ascii', 'replace'))
        if not display_text:
            self.off()
        self.text_idx += 2

    def handle_blink(self):
        if not self.blink:
            return
        self.blink_idx += 1
        if len(self.blink) <= self.blink_idx:
            self.blink_idx = 0
        self.set_backlight_color(self.blink[self.blink_idx])

    def send_manager_data(self, msg, data=None):
        if not data:
            data = []
        if not isinstance(data, (list, tuple)):
            data = [data, ]
        self.pipe.send((
            msg, data
        ))

    def send_raw(self, cmd):
        with open(self.device_path, 'w') as dev:
            dev.write(cmd)

    def send(self, cmd):
        try:
            cmd = cmd + '\n'
            logger.debug(
                'Sending command: "%s"' % cmd.encode('string-escape')
            )
            self.send_raw(cmd)
        except IOError:
            logger.error(
                'Device unavailable; command \'%s\' dropped.',
                cmd
            )

    def get_message_lines(self, message):
        lines = []
        original_lines = message.split('\r')
        for line in original_lines:
            lines.extend(
                line[i:i+self.size[0]]
                for i in range(0, len(line), self.size[0])
            )
        logger.debug(lines)
        return lines

    @command
    def set_contrast(self, value):
        logger.debug('Setting contrast to %s', value)
        self.send('\xfe\x50%s' % chr(value))

    @command
    def set_brightness(self, value):
        logger.debug('Setting brightness to %s', value)
        self.send('\xfe\x99%s' % chr(value))

    @command
    def message(self, message):
        if 'message' in message and self.message != message['message']:
            self.set_message(message['message'])
        if 'blink' in message and self.blink != message['blink']:
            self.set_blink(message['blink'])
        if not 'blink' in message:
            self.set_blink([])
        if not self.blink and message['color'] != self.color:
            self.set_backlight_color(message['color'])

        if message['backlight']:
            self.on()
        else:
            self.off()

    @command
    def set_blink(self, colors):
        self.blink = colors
        self.blink_idx = 0
        if self.blink:
            self.set_backlight_color(self.blink[self.blink_idx])

    @command
    def set_message(self, message):
        logger.debug('Setting message \'%s\'', message)
        self.clear()
        self.text_idx = 0
        self.message = message.replace('\n', '')
        self.message_lines = self.get_message_lines(self.message)
        self.handle_text_cycle()

    @command
    def off(self, *args):
        logger.debug('Setting backlight to off')
        self.send('\xfe\x46')

    @command
    def on(self, *args):
        logger.debug('Setting backlight to on')
        self.send('\xfe\x42')

    @command
    def clear(self, *args):
        self.message = ''
        self.text_idx = 0
        self.message_lines = []
        self.send('\xfe\x58')

    @command
    def set_backlight_color(self, color):
        logger.debug('Setting backlight color to %s', color)
        self.color = color
        self.send('\xfe\xd0%s%s%s' % tuple([chr(c) for c in color]))



message_schema = {
    'title': 'Message Schema',
    'type': 'object',
    'properties': {
        'message': {
            'type': 'string'
        },
        'expires': {
            'oneOf': [
                {
                    # String datetime; parsed by dateutil
                    'type': 'string',
                },
                {
                    # Seconds from now
                    'type': 'integer',
                    'minimum': 0,
                }
            ]
        },
        'color': {
            'type': 'array',
            'minItems': 3,
            'maxItems': 3,
            'items': {
                'type': 'integer',
                'minimum': 0,
                'maximum': 255,
            }
        },
        'blink': {
            'type': 'array',
            'minItems': 1,
            'items': {
                'type': 'array',
                'minItems': 3,
                'maxItems': 3,
                'items': {
                    'type': 'integer',
                    'minimum': 0,
                    'maximum': 255,
                }
            }
        },
        'backlight': {
            'type': 'boolean',
        },
        'timeout': {
            'type': 'integer',
            'minimum': 1,
        },
        'id': {
            'type': 'string',
        }
    },
    'additionalProperties': False,
    'required': ['message']
}

integer_schema = {
    'title': 'Brightness/Contrast Schema',
    'type': 'integer',
    'minimum': 1,
    'maximum': 255
}



import os
import multiprocessing

from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='twoline',
    version='0.7.15',
    url='http://github.com/latestrevision/twoline/',
    description='Manager for simple character LCDs',
    author='Adam Coddington',
    author_email='me@adamcoddington.net',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ],
    install_requires=required,
    packages=find_packages(),
    entry_points={'console_scripts': [
        'twoline = twoline.cmdline:run_from_cmdline']},
    test_suite='nose.collector',
    tests_require=[
        'nose',
    ]
)



import datetime
import json
import logging

from flask import Flask, make_response, request

from twoline.exceptions import InvalidRequest, NotFound, BadRequest


logger = logging.getLogger(__name__)


app = Flask(__name__)


def pipe():
    return app.config['PIPE']


def send_data(msg, data=None):
    if not data:
        data = []
    if not isinstance(data, (list, tuple)):
        data = [data, ]
    logger.debug(
        'Sending data %s%s',
        msg,
        data
    )
    pipe().send((
        msg, data
    ))


def send_and_receive(msg, data=None):
    send_data(msg, data)
    while not pipe().poll():
        pass
    t, args = pipe().recv()
    if t == 'error':
        logger.error('Received error response %s', args)
        raise args[0]
    logger.debug('Received response %s', args)
    return args


def json_response(status_code=200, **kwargs):
    def handle_data(obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        raise TypeError
    response = make_response(
        json.dumps(
            kwargs,
            default=handle_data,
            indent=2
        ),
        status_code
    )
    response.status_code = status_code
    response.headers['Content-Type'] = 'application/json'
    return response


@app.errorhandler(Exception)
def exception_handler(e):
    status_code = 500
    if isinstance(e, InvalidRequest):
        status_code = 422
    elif isinstance(e, NotFound):
        status_code = 404
    elif isinstance(e, BadRequest):
        status_code = 400
    return json_response(
        status_code=status_code,
        error=str(e)
    )


for error in range(400, 599):
    @app.errorhandler(error)
    def error(e):
        return json_response(
            status_code=e.code,
            error=str(e)
        )


@app.route('/', methods=['GET'])
def index():
    routes = {}
    for rule in app.url_map.iter_rules():
        if rule.endpoint in ('index', 'static'):
            continue
        routes[rule.endpoint] = {
            'url': rule.rule,
            'methods': list(rule.methods),
        }
    return json_response(
        **routes
    )


@app.route('/contrast/', methods=['PUT'])
def contrast():
    response = send_and_receive(
        'set_contrast', request.data
    )
    return json_response(
        contrast=response[0]
    )


@app.route('/brightness/', methods=['PUT'])
def brightness():
    response = send_and_receive(
        'set_brightness', request.data
    )
    return json_response(
        brightness=response[0]
    )


@app.route('/message/', methods=['GET', 'POST'])
def message_list():
    if request.method == 'POST':
        response = send_and_receive(
            'post_message', request.data
        )
        return json_response(
            status_code=201,
            **response[0]
        )
    elif request.method == 'GET':
        response = send_and_receive(
            'get_messages', request.data
        )
        return json_response(
            messages=response
        )


@app.route('/flash/', methods=['GET', 'PUT', 'DELETE'])
def flash():
    if request.method == 'PUT':
        response = send_and_receive(
            'put_flash', request.data
        )
        return json_response(
            status_code=201,
            **response[0]
        )
    elif request.method == 'DELETE':
        response = send_and_receive(
            'delete_flash'
        )
        return json_response(
            status=response[0]
        )
    elif request.method == 'GET':
        response = send_and_receive(
            'get_flash'
        )
        return json_response(
            **response[0]
        )


@app.route('/message/<message_id>/', methods=['GET', 'PUT', 'DELETE', 'PATCH'])
def message(message_id):
    if request.method == 'GET':
        response = send_and_receive(
            'get_message_by_id', message_id
        )
        return json_response(
            **response[0]
        )
    elif request.method == 'DELETE':
        response = send_and_receive(
            'delete_message_by_id', message_id
        )
        return json_response(
            status=response[0]
        )
    elif request.method == 'PUT':
        response = send_and_receive(
            'put_message_by_id', [message_id, request.data, ]
        )
        return json_response(
            status_code=201,
            **response[0]
        )
    elif request.method == 'PATCH':
        response = send_and_receive(
            'patch_message_by_id', [message_id, request.data, ]
        )
        return json_response(
            **response[0]
        )



import datetime
from functools import wraps
import json
import logging
import multiprocessing
import time
import uuid

from dateutil.parser import parse
from dateutil.tz import tzlocal
from jsonschema import validate, ValidationError
import pytz

from twoline.exceptions import (
    InvalidRequest, NotFound, BadRequest, UnexpectedError
)
from twoline.lcd import LcdManager
from twoline.schema import message_schema, integer_schema
from twoline.web import app


logger = logging.getLogger(__name__)


WEB_COMMANDS = {}
LCD_COMMANDS = {}


def web_command(fn):
    @wraps(fn)
    def wrapped(*args):
        self = args[0]
        logger.debug(
            'Executing %s%s',
            fn.func_name,
            args
        )
        try:
            response = fn(*args)
            logger.debug(
                'Response %s',
                response
            )
            if response is not None:
                self.send_web_data('response', response)
        except NotFound as e:
            self.send_web_data('error', e)
        except ValidationError as e:
            self.send_web_data('error', InvalidRequest(str(e)))
        except ValueError as e:
            self.send_web_data('error', BadRequest(str(e)))
        except Exception as e:
            self.send_web_data('error', UnexpectedError(str(e)))

    WEB_COMMANDS[fn.func_name] = wrapped
    return wrapped


def lcd_command(fn):
    @wraps(fn)
    def wrapped(*args):
        self = args[0]
        logger.debug(
            'Executing %s%s',
            fn.func_name,
            args
        )
        response = fn(*args)
        logger.debug(
            'Response %s',
            response
        )
        if response is not None:
            self.send_lcd_data('response', response)

    LCD_COMMANDS[fn.func_name] = wrapped
    return fn


class Manager(object):
    def __init__(
        self, device, ip='0.0.0.0', port=9101,
        size_x=16, size_y=2, blink_interval=0.25, text_cycle_interval=2,
        *args, **kwargs
    ):
        self.ip = ip
        self.port = port
        self.device = device
        self.size_x = int(size_x)
        self.size_y = int(size_y)
        self.blink_interval = float(blink_interval)
        self.text_cycle_interval = float(text_cycle_interval)

        self.web_pipe, self.web_proc = self.run_webserver()
        self.lcd_pipe, self.lcd_proc = self.run_lcd()

        self.no_messages = {
            'backlight': False,
            'message': ''
        }
        self.default_message = {
            'color': (255, 255, 255),
            'backlight': True,
            'interval': 5
        }
        self.default_flash = {
            'blink': [(255, 0, 0), (0, 0, 0)],
            'backlight': True,
            'timeout': 10
        }
        self.flash = None
        self.flash_until = None
        self.messages = []
        self._message_id = None
        self.until = None

        self.sleep = 0.2

    @property
    def message_id(self):
        return self._message_id

    @message_id.setter
    def message_id(self, value):
        logger.debug('Setting message_id to %s', self._message_id)
        self._message_id = value

    def run(self):
        logger.info(
            'Listening on http://%s:%s',
            self.ip,
            self.port
        )
        try:
            self._run()
        except Exception as e:
            logger.exception(e)
            try:
                self.lcd_proc.terminate()
            except Exception as e:
                logger.exception(e)
            try:
                self.web_proc.terminate()
            except Exception as e:
                logger.exception(e)

    def _run(self):
        logger.debug("Waiting for data")
        while True:
            if self.web_pipe.poll():
                cmd, args = self.web_pipe.recv()
                args.insert(0, self)
                logger.debug(
                    "Data received from WEB %s:%s",
                    cmd,
                    args
                )
                if cmd in WEB_COMMANDS:
                    logger.info(
                        'WEB Command Received %s%s',
                        cmd,
                        args
                    )
                    WEB_COMMANDS[cmd](*args)
                else:
                    logger.error(
                        'Received unknown command \'%s\' from web.',
                        cmd
                    )
                    self.send_web_data(
                        'error', 'Command %s does not exist' % cmd
                    )
            if self.lcd_pipe.poll():
                cmd, args = self.lcd_pipe.recv()
                args.insert(0, self)
                logger.debug(
                    "Data received from LCD %s:%s",
                    cmd,
                    args
                )
                if cmd in LCD_COMMANDS:
                    logger.info(
                        'LCD Command Received %s%s',
                        cmd,
                        args
                    )
                    WEB_COMMANDS[cmd](*args)
                else:
                    logger.error(
                        'Received unknown command \'%s\' from lcd.',
                        cmd
                    )
                    self.send_lcd_data(
                        'error', 'Command %s does not exist' % cmd
                    )
            time.sleep(self.sleep)
            self.update_screen()

    def get_flash_message(self):
        original_message = self.flash
        logger.debug("Original Flash: %s", original_message)
        default = self.default_flash.copy()
        default.update(
            original_message
        )
        logger.debug("Final Flash: %s", default)
        return default

    def get_message(self):
        if not self.message_id and self.messages:
            self.message_id = self.messages[0]['id']
        idx = self.get_message_index_by_id(self.message_id)
        original_message = self.messages[idx]
        logger.debug("Original Message: %s", original_message)
        default = self.default_message.copy()
        default.update(
            original_message
        )
        logger.debug("Final Message: %s", default)
        return default

    def get_no_messages_message(self):
        message = self.no_messages
        default = self.default_message.copy()
        default.update(
            message
        )
        return default

    def get_message_index_by_id(self, id_):
        for message in self.messages:
            if message['id'] == id_:
                return self.messages.index(message)

    def increment_index(self):
        if self.message_id:
            current_index = self.get_message_index_by_id(self.message_id) + 1
            if len(self.messages) <= current_index:
                current_index = 0
        else:
            current_index = 0
        self.until = None
        self.message_id = self.messages[current_index]['id']
        logger.debug(
            'Incrementing: Index: %s: %s',
            current_index,
            self.message_id
        )

    def delete_message(self, message_id):
        if self.message_id == message_id:
            self.increment_index()
            if self.message_id == message_id:
                # If we just incremented, and are still at the same index
                # there is only one message in the list, and we're
                # in the process of deleting it.
                self.until = None
                self.message_id = None
        idx = self.get_message_index_by_id(message_id)
        message = self.messages[idx]
        self.messages.remove(message)

    def handle_expirations(self):
        utcnow = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
        for message in self.messages:
            if 'expires' in message and message['expires'] < utcnow:
                logger.info(
                    'Message %s has expired.',
                    message['id']
                )
                self.delete_message(message['id'])
        logger.debug('Flash Until: %s', self.flash_until)
        logger.debug('Message Until: %s', self.until)
        if self.flash and self.flash_until and self.flash_until < utcnow:
            logger.info('Flash message has expired')
            self.flash = None
            self.flash_until = None
        if self.messages:
            if self.until and self.until < utcnow:
                self.increment_index()

    def get_current_message(self):
        utcnow = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
        self.handle_expirations()
        if self.flash:
            flash = self.get_flash_message()
            if not self.flash_until:
                self.flash_until = (
                    utcnow + datetime.timedelta(seconds=flash['timeout'])
                )
            return flash
        elif self.messages:
            message = self.get_message()
            if not self.until:
                self.until = (
                    utcnow + datetime.timedelta(seconds=message['interval'])
                )
            return message
        else:
            return self.get_no_messages_message()

    def update_screen(self):
        try:
            message = self.get_current_message()
        except TypeError:
            self.message_id = None
            message = None

        if not message:
            logger.warning(
                "Message was lost before it could be displayed; skipping."
            )
            return

        self.send_lcd_data(
            'message', message
        )

    def send_lcd_data(self, msg, data=None):
        if not data:
            data = []
        if not isinstance(data, (list, tuple)):
            data = [data, ]
        self.lcd_pipe.send((
            msg, data
        ))

    def send_web_data(self, msg, data=None):
        if not data:
            data = []
        if not isinstance(data, (list, tuple)):
            data = [data, ]
        self.web_pipe.send((
            msg, data
        ))

    def run_lcd(self):
        local, lcd_pipe = multiprocessing.Pipe()

        def _run_lcd():
            mgr = LcdManager(
                self.device,
                lcd_pipe,
                size_x=self.size_x,
                size_y=self.size_y,
                blink_interval=self.blink_interval,
                text_cycle_interval=self.text_cycle_interval
            )
            mgr.initialize()
            mgr.run()

        process = multiprocessing.Process(
            target=_run_lcd
        )
        process.start()
        logger.debug(
            'Started LCD on pid %s',
            process.pid
        )
        return local, process

    def run_webserver(self):
        local, webserver = multiprocessing.Pipe()

        def _run_webserver():
            app.config['PIPE'] = webserver
            app.run(
                host=self.ip,
                port=int(self.port),
            )
        process = multiprocessing.Process(
            target=_run_webserver
        )
        process.start()
        logger.debug(
            'Started WEB on pid %s',
            process.pid
        )
        return local, process

    def process_message(self, message, ignore_id=False):
        if 'expires' in message:
            if isinstance(message['expires'], datetime.datetime):
                message['expires'] = message['expires'].isoformat()
        validate(message, message_schema)
        if not 'id' in message or ignore_id:
            message['id'] = uuid.uuid4().hex
        if 'expires' in message:
            if isinstance(message['expires'], int):
                message['expires'] = (
                    datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
                    + datetime.timedelta(seconds=message['expires'])
                )
            elif isinstance(message['expires'], basestring):
                try:
                    expires = parse(message['expires'])
                    if not expires.tzinfo:
                        expires = expires.replace(tzinfo=tzlocal())
                    message['expires'] = expires
                except ValueError:
                    raise ValidationError(
                        (
                            '\'%s\' is neither a valid ISO datetime or an '
                            'integer count of seconds'
                        ) % message['expires']
                    )
            else:
                raise ValidationError(
                    (
                        '\'%s\' is neither a valid ISO datetime or an integer '
                        'count of seconds'
                    ) % message['expires']
                )
        return message

    @web_command
    def get_message_by_id(self, id_):
        idx = self.get_message_index_by_id(id_)
        if idx is None:
            raise NotFound('Message %s does not exist' % id_)
        return self.messages[idx]

    @web_command
    def delete_message_by_id(self, id_):
        idx = self.get_message_index_by_id(id_)
        if idx is None:
            raise NotFound('Message %s does not exist' % id_)
        self.delete_message(id_)
        return 'OK'

    @web_command
    def put_message_by_id(self, id_, message_payload):
        message = json.loads(message_payload)
        idx = self.get_message_index_by_id(id_)
        if idx is None:
            message['id'] = id_
            self.messages.append(self.process_message(message))
            return message
        else:
            self.messages[idx] = self.process_message(message)
            self.messages[idx]['id'] = id_
            return self.messages[idx]

    @web_command
    def patch_message_by_id(self, id_, message_payload):
        message = json.loads(message_payload)
        idx = self.get_message_index_by_id(id_)
        if idx is None:
            raise NotFound('Message %s does not exist' % id_)
        original_message = self.messages[idx]
        original_message.update(message)
        self.messages[idx] = self.process_message(original_message)
        return self.messages[idx]

    @web_command
    def set_brightness(self, value):
        try:
            value = int(value)
        except ValueError:
            raise ValidationError('Brightness requires an integer value')
        validate(value, integer_schema)
        self.send_lcd_data(
            'set_brightness', int(value)
        )
        return value

    @web_command
    def set_contrast(self, value):
        try:
            value = int(value)
        except ValueError:
            raise ValidationError('Contrast requires an integer value')
        validate(value, integer_schema)
        self.send_lcd_data(
            'set_contrast', int(value)
        )
        return value

    @web_command
    def get_messages(self, *args):
        return self.messages

    @web_command
    def post_message(self, message_payload):
        message = json.loads(message_payload)
        self.messages.append(
            self.process_message(
                message,
                ignore_id=True
            )
        )
        return message

    @web_command
    def put_flash(self, message_payload):
        self.flash = self.process_message(
            json.loads(
                message_payload
            )
        )
        return self.get_flash_message()  # Post-processing

    @web_command
    def delete_flash(self):
        self.flash = None
        self.flash_until = None
        return 'OK'

    @web_command
    def get_flash(self):
        if not self.flash:
            raise NotFound('Flash message not set')
        return self.flash

    @lcd_command
    @web_command
    def error(self, *args):
        print 'Error: %s' % [args]
